# MIT License
#
# Copyright (c) 2021 TrigonDev
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    Iterable,
    Type,
    TypeVar,
    cast,
    overload,
)

if TYPE_CHECKING:  # pragma: no cover
    from .connection import Connection
    from .field import BaseField
    from .model import Model
    from .utils.lazy_list import LazyList


_REF = TypeVar("_REF", bound="Model")
_THROUGH = TypeVar("_THROUGH", bound="Model")


class ManyToMany(Generic[_REF, _THROUGH]):
    """A useful tool to simplify many-to-many references.

    Args:
        here (str): The field name on the current model.
        here_ref (str): The model and field name on the "middle" table (in the
        example below, the middle table is Player) in the format of
        "model.field".
        other_ref (str): The model and field name on the middle table that
        references the final table (or other table).
        other (str): The model and field name on the final table referenced by
        the middle table.

    Note: Although unecessary, it is highly recommended to use ForeignKeys on
    the middle table where it references the initial and final table. You may
    get unexpected behaviour if you don't.

    Example Usage:
    ```
    class User(Model):
        username = VarChar(32).field()
        primary_key = (username,)

        games = ManyToMany["Game", "Player"](  # the typehints are optional
            # the column on this table referenced in Player
            "username",

            # the column on Player that references "username"
            "players.username",

            # the column on Player that references Game.gameid
            "players.gameid",

            # the column on Game referenced by Player
            "games.gameid",
        )

    class Game(Model):
        gameid = Serial().field()
        primary_key = (gameid,)

        users = ManyToMany["User", "Player"](
            "gameid",
            "players.gameid",
            "players.username",
            "users.username",
        )

    class Player(Model):
        username = VarChar(32).field()
        gameid = Int().field()
        primary_key = (username, gameid)

        username_fk = ForeignKey(username, User.username)
        gameid_fk = ForeignKey(gameid, Game.gameid)

    class MyDatabase(Database):
        users = User
        games = Game
        players = Player

    ...

    circuit = await User.fetch(username="Circuit")
    circuits_games = await circuit.games.fetchmany()
    ```

    If you want typehints to work properly, use
    `games = ManyToMany["Game"](...)`.
    """

    __slots__: Iterable[str] = (
        "_here",
        "_here_ref",
        "_other_ref",
        "_other",
        "_attribute_name",
    )

    _attribute_name: str
    # populated by Model on __init_subclass__

    def __init__(
        self, here: str, here_ref: str, other_ref: str, other: str
    ) -> None:
        self._here = here
        self._here_ref = here_ref
        self._other_ref = other_ref
        self._other = other

    @overload
    def __get__(
        self, inst: Model, cls: Type[Model]
    ) -> _RealManyToMany[_REF, _THROUGH]:
        ...

    @overload
    def __get__(
        self, inst: None, cls: Type[Model]
    ) -> ManyToMany[_REF, _THROUGH]:
        ...

    def __get__(
        self, inst: Model | None, cls: type[Model]
    ) -> ManyToMany[_REF, _THROUGH] | _RealManyToMany[_REF, _THROUGH]:
        if inst is None:
            return self
        real_m2m = self._generate_mtm(inst)
        setattr(inst, self._attribute_name, real_m2m)
        return real_m2m

    def _generate_mtm(self, inst: Model) -> _RealManyToMany[_REF, _THROUGH]:
        return _RealManyToMany(self, inst)


class _RealManyToMany(Generic[_REF, _THROUGH]):
    __slots__: Iterable[str] = (
        "orig",
        "model",
        "field",
        "mm_model",
        "mm_h_field",
        "mm_o_field",
        "ot_model",
        "ot_field",
    )

    def __init__(
        self, orig: ManyToMany[_REF, _THROUGH], model_inst: Model
    ) -> None:
        # NOTE: all these casts are ugly, but truthfully
        # there isn't a better way to do this. You can't
        # actually check that these are Models and Fields
        # without creating circular imports (since model.py
        # imports this file)

        self.orig = orig

        mm_h_model, _mm_h_field = self.orig._here_ref.split(".")
        mm_o_model, _mm_o_field = self.orig._other_ref.split(".")
        assert mm_h_model == mm_o_model

        mm_model = cast(
            "Type[Model]", getattr(model_inst.database, mm_h_model)
        )

        mm_h_field = cast(
            "BaseField[Any, Any, Any]", getattr(mm_model, _mm_h_field)
        )
        mm_o_field = cast(
            "BaseField[Any, Any, Any]", getattr(mm_model, _mm_o_field)
        )

        _ot_model, _ot_field = self.orig._other.split(".")
        ot_model = cast("Type[Model]", getattr(model_inst.database, _ot_model))

        ot_field = cast(
            "BaseField[Any, Any, Any]", getattr(ot_model, _ot_field)
        )

        self.model = model_inst
        self.field = cast(
            "BaseField[Any, Any, Any]",
            getattr(model_inst.__class__, self.orig._here),
        )
        self.mm_model = mm_model
        self.mm_h_field = mm_h_field
        self.mm_o_field = mm_o_field
        self.ot_model = ot_model
        self.ot_field = ot_field

    def __getattr__(self, name: str) -> Any:
        return getattr(self.orig, name)

    async def fetchmany(
        self, con: Connection | None = None
    ) -> LazyList[dict[str, Any], Model]:
        """Fetch all rows from the final table that belong to this instance.

        Returns:
            LazyList[dict, Model]: A lazy-list of returned Models.
        """

        return (
            await self.ot_model.fetch_query(con=con)
            .where(
                self.mm_model.fetch_query()
                .where(
                    self.mm_h_field.eq(
                        self.model._raw_values[self.field.name]
                    ),
                    self.mm_o_field.eq(self.ot_field),
                )
                .exists()
            )
            .fetchmany()
        )

    async def count(self, con: Connection | None = None) -> int:
        """Returns the count.

        Warning: To be efficient, this returns the count of *middle* models,
        which may differ from the number of final models if you did not use
        ForeignKeys properly.

        Returns:
            int: The count.
        """

        return (
            await self.mm_model.fetch_query(con=con)
            .where(self.mm_h_field.eq(self.model._raw_values[self.field.name]))
            .count()
        )

    async def clear(
        self, con: Connection | None = None
    ) -> LazyList[dict[str, Any], Model]:
        """Remove all instances of the other model from this instance.

        Both of these lines do the same thing:
        ```
        deleted_players = await user.games.clear()
        deleted_players = await Player.delete_query().where(
            username=user.name
        ).execute()
        ```

        Returns:
            LazyList[dict, _REF]: A lazy-list of deleted through models (in
            the example, it would be a list of Player).
        """

        return (
            await self.mm_model.delete_query(con=con)
            .where(self.mm_h_field.eq(self.model._raw_values[self.field.name]))
            .execute()
        )

    async def add(self, other: Model, con: Connection | None = None) -> Model:
        """Add one or more models to this ManyToMany.

        Each of these lines does the exact same thing:
        ```
        player = await user.games.add(game)
        # OR
        player = await games.users.add(user)
        # OR
        player = await Player(username=user.name, gameid=game.id).create()
        ```

        Returns:
            Model: The reference model that lines this model and the other
            model. In the example, the return would be a Player.
        """

        values = {
            self.mm_h_field.name: self.model._raw_values[self.field.name],
            self.mm_o_field.name: other._raw_values[self.ot_field.name],
        }
        return await self.mm_model(**values).create(con=con)

    async def remove(
        self, other: Model, con: Connection | None = None
    ) -> LazyList[dict[str, Any], Model]:
        """Remove one or models from this ManyToMany.

        Each of these lines does the exact same thing:
        ```
        deleted_players = await user.games.remove(game)
        # OR
        deleted_players = await games.user.remove(user)
        # OR
        deleted_players = await Player.delete_query().where(
            username=user.name, gameid=game.id
        ).execute()
        ```

        Note: The fact that .remove() returns a list instead of a single model
        was intentional. The reason is ManyToMany does not enforce uniqueness
        in any way, so there could be multiple Players that link a single user
        to a single game. Thus, user.remove(game) could actually end up
        deleting multiple players.
        """

        values = {
            self.mm_h_field.name: self.model._raw_values[self.field.name],
            self.mm_o_field.name: other._raw_values[self.ot_field.name],
        }
        return (
            await self.mm_model.delete_query(con=con).where(**values).execute()
        )



# MIT License
#
# Copyright (c) 2021 TrigonDev
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from enum import Enum
from typing import Any, Iterable, Sequence
from typing import cast as typingcast

from apgorm.exceptions import BadArgument
from apgorm.field import BaseField
from apgorm.sql.sql import Block, join, raw

from .constraint import Constraint


class ForeignKeyAction(Enum):
    """Action for ON UPDATE or ON DELETE of ForeignKey."""

    CASCADE = "CASCADE"
    """Carry the changes"""
    RESTRICT = "RESTRICT"
    """Prevent the changes"""
    NO_ACTION = "NO ACTION"
    """Do nothing"""


class ForeignKey(Constraint):
    __slots__: Iterable[str] = (
        "fields",
        "ref_fields",
        "ref_table",
        "match_full",
        "on_delete",
        "on_update",
    )

    def __init__(
        self,
        fields: Sequence[Block[Any] | BaseField[Any, Any, Any] | str]
        | Block[Any]
        | BaseField[Any, Any, Any]
        | str,
        ref_fields: Sequence[Block[Any] | BaseField[Any, Any, Any] | str]
        | Block[Any]
        | BaseField[Any, Any, Any]
        | str,
        ref_table: Block[Any] | str | None = None,
        match_full: bool = False,
        on_delete: ForeignKeyAction = ForeignKeyAction.CASCADE,
        on_update: ForeignKeyAction = ForeignKeyAction.CASCADE,
    ) -> None:
        """Specify a ForeignKey constraint for a table.

        Args:
            fields (Sequence[Block | BaseField]): A list of fields or raw
            field names on the current table.
            ref_fields (Sequence[Block | BaseField]): A list of fields or raw
            field names on the referenced table.
            ref_table (Block, optional): If all of `ref_fields` are `Block`,
            specify the raw tablename of the referenced table. Defaults to
            None.
            match_full (bool, optional): Whether or not a full match is
            required. If False, MATCH SIMPLE is used instead. Defaults to
            False.
            on_delete (Action, optional): The action to perform if the
            referenced row is deleted. Defaults to Action.CASCADE.
            on_update (Action, optional): The action to perform if the
            referenced row is updated. Defaults to Action.CASCADE.

        Raises:
            BadArgument: Bad arguments were sent to ForeignKey.
        """

        self.fields: Sequence[BaseField[Any, Any, Any] | Block[Any]] = [
            raw(f) if isinstance(f, str) else f
            for f in (fields if isinstance(fields, Sequence) else [fields])
        ]
        self.ref_fields: Sequence[BaseField[Any, Any, Any] | Block[Any]] = [
            raw(f) if isinstance(f, str) else f
            for f in (
                ref_fields
                if isinstance(ref_fields, Sequence)
                else [ref_fields]
            )
        ]
        self.ref_table = (
            raw(ref_table) if isinstance(ref_table, str) else ref_table
        )
        self.match_full = match_full
        self.on_delete = on_delete
        self.on_update = on_update

        if len(self.ref_fields) != len(self.fields):
            raise BadArgument(
                "Must have same number of fields and ref_fields."
            )

        if not self.fields:
            raise BadArgument("Must specify at least on field and ref_field.")

    def _creation_sql(self) -> Block[Any]:

        if (
            len(
                {
                    f.model.tablename
                    for f in self.ref_fields
                    if isinstance(f, BaseField)
                }
            )
            > 1
        ):
            raise BadArgument(
                "All fields in ref_fields must be of the same table."
            )

        if self.ref_table is None:
            _ref_fields = self.ref_fields
            if not all(isinstance(f, BaseField) for f in _ref_fields):
                raise BadArgument(
                    "ref_fields must either all be BaseFields or "
                    "ref_table must be specified."
                )
            _ref_fields = typingcast(
                Sequence[BaseField[Any, Any, Any]], _ref_fields
            )

            ref_table = raw(_ref_fields[0].model.tablename)

        else:
            ref_table = self.ref_table

        ref_fields = (
            raw(f.name) if isinstance(f, BaseField) else f
            for f in self.ref_fields
        )
        fields = (
            raw(f.name) if isinstance(f, BaseField) else f for f in self.fields
        )

        return Block(
            raw("CONSTRAINT"),
            raw(self.name),
            raw("FOREIGN KEY ("),
            join(raw(","), *fields),
            raw(") REFERENCES"),
            ref_table,
            raw("("),
            join(raw(","), *ref_fields),
            raw(") MATCH"),
            raw("FULL" if self.match_full else "SIMPLE"),
            raw(f"ON DELETE {self.on_delete.value}"),
            raw(f"ON UPDATE {self.on_update.value}"),
            wrap=True,
        )



# MIT License
#
# Copyright (c) 2021 TrigonDev
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from decimal import Decimal
from typing import Iterable, TypeVar, Union

from apgorm.exceptions import BadArgument
from apgorm.field import Field

from .base_type import SqlType

INT = Union["SmallInt", "Int", "BigInt"]
"""All integral types (excluding serials)."""
FLOAT = Union["Real", "DoublePrecision", "Numeric"]
"""All floating-point types."""
SERIAL = Union["SmallSerial", "Serial", "BigSerial"]
"""All serial types."""

NUMBER = Union[INT, FLOAT, SERIAL]
"""All integral, floating-point, and serial types."""


class SmallInt(SqlType[int]):
    """Small integer, 2 bytes, -32768 to +32767.

    https://www.postgresql.org/docs/14/datatype-numeric.html
    """

    __slots__: Iterable[str] = ()

    _sql = "SMALLINT"


class Int(SqlType[int]):
    """Integer, 4 bytes, -2147483648 to +2147483647.

    https://www.postgresql.org/docs/14/datatype-numeric.html
    """

    __slots__: Iterable[str] = ()

    _sql = "INTEGER"


Integer = Int


class BigInt(SqlType[int]):
    """Big integer, 8 bytes, -9223372036854775808 to +9223372036854775807.

    https://www.postgresql.org/docs/14/datatype-numeric.html
    """

    __slots__: Iterable[str] = ()

    _sql = "BIGINT"


class Numeric(SqlType[Decimal]):
    """Numeric (AKA decimal) type with a user specified precision and scale.

    Up to 131072 digits before the decimal point and 16383 digits after.

    Args:
        precision (int, optional): The total number of significant digits
        (`55.55` has 4). Defaults to None.
        scale (int, optional): The total number of digits after the
        decimal. Specifying a precision without a scale sets the scale to
        0. Defaults to None.

    Raises:
        BadArgument: You tried to specify a scale without a precision.

    https://www.postgresql.org/docs/14/datatype-numeric.html
    """

    __slots__: Iterable[str] = ("_precision", "_scale")

    def __init__(
        self, precision: int | None = None, scale: int | None = None
    ) -> None:
        self._precision = precision
        self._scale = scale

        self._sql = "NUMERIC"

        if precision is not None and scale is not None:
            self._sql += f"({precision}, {scale})"
        elif precision is not None:
            self._sql += f"({precision})"
        elif scale is not None:
            raise BadArgument("Cannot specify scale without precision.")

    @property
    def precision(self) -> int | None:
        """The precision (total number of significant digits) of the type.

        Returns:
            int | None
        """

        return self._precision

    @property
    def scale(self) -> int | None:
        """The scale (digits after the decimal point) of the type.

        Returns:
            int | None
        """

        return self._scale


class Real(SqlType[float]):
    """4 byte floating-point number (inexact).

    https://www.postgresql.org/docs/14/datatype-numeric.html
    """

    __slots__: Iterable[str] = ()

    _sql = "REAL"


class DoublePrecision(SqlType[float]):
    """8 byte floating-point number (inexact).

    https://www.postgresql.org/docs/14/datatype-numeric.html
    """

    __slots__: Iterable[str] = ()

    _sql = "DOUBLE PRECISION"


_S = TypeVar("_S", bound="_BaseSerial", covariant=True)


class _BaseSerial(SqlType[int]):
    __slots__: Iterable[str] = ()

    def field(  # type: ignore
        self: _S, use_repr: bool = True
    ) -> Field[_S, int]:
        return Field(sql_type=self, not_null=True, use_repr=use_repr)

    def nullablefield(  # type: ignore
        self: _S, use_repr: bool = True
    ) -> Field[_S, int | None]:
        return Field(sql_type=self, not_null=False, use_repr=use_repr)


class SmallSerial(_BaseSerial):
    """Small autoincrementing integer. You cannot cast to this type.

    Accepts any value accepted by SmallInt, but the default values are
    positive only (starting at 1).

    https://www.postgresql.org/docs/14/datatype-numeric.html
    """

    __slots__: Iterable[str] = ()

    _sql = "SMALLSERIAL"


class Serial(_BaseSerial):
    """Autoincrementing integer. You cannot cast to this type.

    Accepts any value accepted by Int, but the default values are positive
    only (starting at 1).

    https://www.postgresql.org/docs/14/datatype-numeric.html
    """

    __slots__: Iterable[str] = ()

    _sql = "SERIAL"


class BigSerial(_BaseSerial):
    """Large autoincrementing integer. You cannot cast to this type.

    Accepts any value accepted by BigInt, but the default values are positive
    only (starting at 1).

    https://www.postgresql.org/docs/14/datatype-numeric.html
    """

    __slots__: Iterable[str] = ()

    _sql = "BIGSERIAL"



# MIT License
#
# Copyright (c) 2021 TrigonDev
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from apgorm.sql.generators import alter
from apgorm.sql.sql import raw
from apgorm.undefined import UNDEF

from .describe import DescribeConstraint
from .migration import Migration

if TYPE_CHECKING:  # pragma: no cover
    from .describe import Describe


def _handle_constraint_list(
    tablename: str,
    orig: list[DescribeConstraint],
    curr: list[DescribeConstraint],
) -> tuple[list[str], list[str]]:
    origd = {c.name: c for c in orig}
    currd = {c.name: c for c in curr}

    new_constraints = [
        alter.add_constraint(raw(tablename), c.raw_sql).render_no_params()
        for c in currd.values()
        if c.name not in origd
    ]
    drop_constraints = [
        alter.drop_constraint(raw(tablename), raw(c.name)).render_no_params()
        for c in origd.values()
        if c.name not in currd
    ]

    common = ((currd[k], origd[k]) for k in currd.keys() & origd.keys())
    for currc, origc in common:
        if currc.raw_sql == origc.raw_sql:
            continue

        new_constraints.append(
            alter.add_constraint(
                raw(tablename), currc.raw_sql
            ).render_no_params()
        )
        drop_constraints.append(
            alter.drop_constraint(
                raw(tablename), raw(currc.name)
            ).render_no_params()
        )

    return new_constraints, drop_constraints


def create_next_migration(cd: Describe, folder: Path) -> str | None:
    lm = Migration._load_last_migration(folder)

    curr_tables = {t.name: t for t in cd.tables}
    last_tables = {t.name: t for t in lm.describe.tables} if lm else {}

    migrations: list[str] = []

    # tables  # TODO: renamed tables
    add_tables = [
        alter.add_table(raw(key)).render_no_params()
        for key in curr_tables
        if key not in last_tables
    ]
    drop_tables = [
        alter.drop_table(raw(key)).render_no_params()
        for key in last_tables
        if key not in curr_tables
    ]

    # everything else
    curr_indexes = {c.name: c for c in cd.indexes}
    last_indexes = {c.name: c for c in lm.describe.indexes} if lm else {}
    comm_indexes = (
        (curr_indexes[k], last_indexes[k])
        for k in curr_indexes.keys() & last_indexes.keys()
    )
    add_indexes: list[str] = [
        alter.add_index(c.raw_sql).render_no_params()
        for name, c in curr_indexes.items()
        if name not in last_indexes
    ]
    drop_indexes: list[str] = [
        alter.drop_index(raw(c.name)).render_no_params()
        for name, c in last_indexes.items()
        if name not in curr_indexes
    ]
    for curr, last in comm_indexes:
        if curr.raw_sql != last.raw_sql:
            drop_indexes.append(
                alter.drop_index(raw(curr.name)).render_no_params()
            )
            add_indexes.append(
                alter.add_index(curr.raw_sql).render_no_params()
            )

    add_unique_constraints: list[str] = []
    add_pk_constraints: list[str] = []
    add_check_constraints: list[str] = []
    add_fk_constraints: list[str] = []
    add_exclude_constraints: list[str] = []

    drop_unique_constraints: list[str] = []
    drop_pk_constraints: list[str] = []
    drop_check_constraints: list[str] = []
    drop_fk_constraints: list[str] = []
    drop_exclude_constraints: list[str] = []

    add_fields: list[str] = []
    drop_fields: list[str] = []

    field_not_nulls: list[str] = []

    for tablename, currtable in curr_tables.items():
        lasttable = (
            last_tables[tablename] if tablename in last_tables else None
        )

        # fields:
        curr_fields = {f.name: f for f in currtable.fields}
        last_fields = (
            {f.name: f for f in lasttable.fields} if lasttable else {}
        )

        add_fields.extend(
            alter.add_field(
                raw(tablename), raw(key), raw(f.type_)
            ).render_no_params()
            for key, f in curr_fields.items()
            if key not in last_fields
        )
        drop_fields.extend(
            alter.drop_field(raw(tablename), raw(key)).render_no_params()
            for key, f in last_fields.items()
            if key not in curr_fields
        )

        # field not nulls:
        for field in curr_fields.values():
            last_not_null = (
                UNDEF.UNDEF
                if field.name not in last_fields
                else last_fields[field.name].not_null
            )
            set_nn_to: bool | None = None
            if last_not_null is UNDEF.UNDEF:
                if field.not_null:
                    set_nn_to = True
            elif last_not_null != field.not_null:
                set_nn_to = field.not_null

            if set_nn_to is not None:
                field_not_nulls.append(
                    alter.set_field_not_null(
                        raw(tablename), raw(field.name), set_nn_to
                    ).render_no_params()
                )

        # constraints (in the order they should be applied)
        _new, _drop = _handle_constraint_list(
            tablename,
            lasttable.unique_constraints if lasttable else [],
            currtable.unique_constraints,
        )
        add_unique_constraints.extend(_new)
        drop_unique_constraints.extend(_drop)

        _new, _drop = _handle_constraint_list(
            tablename,
            [lasttable.pk_constraint] if lasttable else [],
            [currtable.pk_constraint],
        )
        add_pk_constraints.extend(_new)
        drop_pk_constraints.extend(_drop)

        _new, _drop = _handle_constraint_list(
            tablename,
            lasttable.check_constraints if lasttable else [],
            currtable.check_constraints,
        )
        add_check_constraints.extend(_new)
        drop_check_constraints.extend(_drop)

        _new, _drop = _handle_constraint_list(
            tablename,
            lasttable.fk_constraints if lasttable else [],
            currtable.fk_constraints,
        )
        add_fk_constraints.extend(_new)
        drop_fk_constraints.extend(_drop)

        _new, _drop = _handle_constraint_list(
            tablename,
            lasttable.exclude_constraints if lasttable else [],
            currtable.exclude_constraints,
        )
        add_exclude_constraints.extend(_new)
        drop_exclude_constraints.extend(_drop)

    for tablename, lasttable in last_tables.items():
        if tablename in curr_tables:
            continue
        _, _drop = _handle_constraint_list(
            tablename, lasttable.fk_constraints, []
        )
        drop_fk_constraints.extend(_drop)

    # finalization
    migrations.extend(add_tables)

    migrations.extend(drop_fk_constraints)

    migrations.extend(drop_indexes)

    migrations.extend(drop_tables)

    migrations.extend(add_fields)

    migrations.extend(drop_pk_constraints)
    migrations.extend(drop_unique_constraints)
    migrations.extend(drop_check_constraints)
    migrations.extend(drop_exclude_constraints)

    migrations.extend(drop_fields)

    migrations.extend(field_not_nulls)

    migrations.extend(add_indexes)

    migrations.extend(add_unique_constraints)
    migrations.extend(add_check_constraints)
    migrations.extend(add_pk_constraints)
    migrations.extend(add_fk_constraints)
    migrations.extend(add_exclude_constraints)

    if not migrations:
        return None
    return ";\n".join(migrations) + ";"



# MIT License
#
# Copyright (c) 2021 TrigonDev
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from typing import Any, Iterable, Optional, Sequence, TypeVar

from .base_type import SqlType

_T = TypeVar("_T")


class Array(SqlType[Sequence[Optional[_T]]]):
    """SQL array type.

    Args:
        subtype (SqlType): The subtype (can be another Array for multiple
        dimensions).

    Example:
    ```
        list_of_names = Array(VarChar(32)).field()
        game_board = Array(Array(Int(), size=16), size=16).nullablefield()
    ```

    https://www.postgresql.org/docs/14/arrays.html
    """

    __slots__: Iterable[str] = ("_subtype",)

    def __init__(self, subtype: SqlType[_T]) -> None:
        self._subtype = subtype

        def _get_arrays(
            t: SqlType[Any], arrays: list[Array[Any]] | None = None
        ) -> tuple[list[Array[Any]], SqlType[Any]]:
            arrays = arrays or []
            if isinstance(t, Array):
                arrays.append(t)
                return _get_arrays(t.subtype, arrays=arrays)
            else:
                return arrays, t

        arrays, final = _get_arrays(self)

        self._sql = final._sql + "".join(["[]" for _ in reversed(arrays)])

    @property
    def subtype(self) -> SqlType[_T]:
        """The arrays sub type.

        Returns:
            SqlType: The sub type.
        """

        return self._subtype



# MIT License
#
# Copyright (c) 2021 TrigonDev
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from .array import Array
from .binary import ByteA
from .bitstr import Bit, VarBit
from .boolean import Bool, Boolean
from .character import Char, Text, VarChar
from .date import (
    Date,
    Interval,
    IntervalField,
    Time,
    Timestamp,
    TimestampTZ,
    TimeTZ,
)
from .geometric import Box, Circle, Line, LineSegment, Path, Point, Polygon
from .json_type import Json, JsonB
from .monetary import Money
from .network import CIDR, INET, MacAddr, MacAddr8
from .numeric import (
    FLOAT,
    INT,
    NUMBER,
    SERIAL,
    BigInt,
    BigSerial,
    DoublePrecision,
    Int,
    Integer,
    Numeric,
    Real,
    Serial,
    SmallInt,
    SmallSerial,
)
from .uuid_type import UUID
from .xml_type import XML

__all__ = (
    "Array",
    "ByteA",
    "Bit",
    "VarBit",
    "Bool",
    "Boolean",
    "Char",
    "Text",
    "VarChar",
    "Date",
    "Interval",
    "IntervalField",
    "Time",
    "Timestamp",
    "TimestampTZ",
    "TimeTZ",
    "Box",
    "Circle",
    "Line",
    "LineSegment",
    "Path",
    "Point",
    "Polygon",
    "Json",
    "JsonB",
    "Money",
    "CIDR",
    "INET",
    "MacAddr",
    "MacAddr8",
    "FLOAT",
    "INT",
    "NUMBER",
    "SERIAL",
    "BigInt",
    "BigSerial",
    "DoublePrecision",
    "Int",
    "Integer",
    "Numeric",
    "Real",
    "Serial",
    "SmallInt",
    "SmallSerial",
    "UUID",
    "XML",
)



# MIT License
#
# Copyright (c) 2021 TrigonDev
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from enum import IntEnum, IntFlag
from typing import Generic, Iterable, Type, TypeVar, Union, cast

_ORIG = TypeVar("_ORIG")
_CONV = TypeVar("_CONV")


class Converter(Generic[_ORIG, _CONV]):
    """Base class that must be used for all field converters."""

    __slots__: Iterable[str] = ()

    def from_stored(self, value: _ORIG) -> _CONV:
        """Take the value given by the database and convert it to the type
        used in your code."""

        raise NotImplementedError  # pragma: no cover

    def to_stored(self, value: _CONV) -> _ORIG:
        """Take the type used by your code and convert it to the type
        used to store the value in the database."""

        raise NotImplementedError  # pragma: no cover


_INTEF = TypeVar("_INTEF", bound="Union[IntFlag, IntEnum]")


class IntEFConverter(Converter[int, _INTEF], Generic[_INTEF]):
    """Converter that converts integers to IntEnums or IntFlags.
    ```
    class User(Model):
        flags = Int().field().with_converter(IntEFConverter(MyIntFlag))
    ```

    Args:
        type_ (Type[IntEnum] | Type[IntFlag]): The IntEnum or IntFlag to use.
    """

    __slots__: Iterable[str] = ("_type",)

    def __init__(self, type_: Type[_INTEF]):
        self._type: Type[_INTEF] = type_

    def from_stored(self, value: int) -> _INTEF:
        return cast(_INTEF, self._type(value))

    def to_stored(self, value: _INTEF) -> int:
        return cast(Union[IntEnum, IntFlag], value).value



# MIT License
#
# Copyright (c) 2021 TrigonDev
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Any,
    AsyncGenerator,
    Callable,
    Generic,
    Iterable,
    Type,
    TypeVar,
    cast,
)

from apgorm.connection import Connection
from apgorm.undefined import UNDEF
from apgorm.utils.lazy_list import LazyList

from .generators.query import delete, insert, select, update
from .sql import SQL, Block, and_, raw, sql, wrap

if TYPE_CHECKING:  # pragma: no cover
    from apgorm.model import Model
    from apgorm.types.boolean import Bool

_T = TypeVar("_T", bound="Model")


def _dict_model_converter(model: Type[_T]) -> Callable[[dict[str, Any]], _T]:
    def converter(values: dict[str, Any]) -> _T:
        return model._from_raw(**values)

    return converter


class BaseQueryBuilder(Generic[_T]):
    """Base class for query builders."""

    __slots__: Iterable[str] = ("model", "con")

    def __init__(self, model: Type[_T], con: Connection | None = None) -> None:
        self.model = model
        self.con = con or model.database

    def _get_block(self) -> Block[Any]:
        """Convert the data in the query builder to a Block."""

        raise NotImplementedError  # pragma: no cover


_S = TypeVar("_S", bound="FilterQueryBuilder[Any]")


class FilterQueryBuilder(BaseQueryBuilder[_T]):
    """Base class for query builders that have "where logic"."""

    __slots__: Iterable[str] = ("_filters",)

    def __init__(self, model: Type[_T], con: Connection | None = None) -> None:
        super().__init__(model, con)

        self._filters: list[Block[Bool]] = []

    def where(self: _S, *filters: Block[Bool], **values: SQL[Any]) -> _S:
        """Extend the current where logic.

        Example:
        ```
        builder.where(User.nickname.neq("Nickname"), status=4)
        # nickname != "Nickname", status == 4
        ```

        Returns:
            FilterQueryBuilder: Returns the query builder to allow for
            chaining.
        """

        # NOTE: Although **values might look like an SQL-injection
        # vulnerability, it's really not. Since the keys for **values
        # can only contain A-Za-z_ characters, there's no possibly way
        # to perform sql injection, even if the keys are user input.
        self._filters.extend(filters)
        for k, v in values.items():
            self._filters.append(raw(k).eq(v))

        return self

    def _where_logic(self) -> Block[Bool] | None:
        if not self._filters:
            return None
        return and_(*self._filters)


class FetchQueryBuilder(FilterQueryBuilder[_T]):
    """Query builder for fetching models."""

    __slots__: Iterable[str] = ("_order_by_logic", "_reverse")

    def __init__(self, model: Type[_T], con: Connection | None = None) -> None:
        super().__init__(model, con)

        self._order_by_logic: SQL[Any] | UNDEF = UNDEF.UNDEF
        self._reverse: bool = False

    def order_by(
        self, logic: SQL[Any], reverse: bool = False
    ) -> FetchQueryBuilder[_T]:
        """Specify the order logic of the query.

        Args:
            field (Block | BaseField): The field or raw field name to order by.
            reverse (bool, optional): If set, will return results decending
            instead of ascending.

        Returns:
            FetchQueryBuilder: Returns the query builder to allow for chaining.
        """

        self._order_by_logic = logic
        self._reverse = reverse
        return self

    def exists(self) -> Block[Bool]:
        """Returns this query wrapped in EXISTS (). Useful for subqueries:

        ```
        user = await User.fetch(name="Circuit")
        games = await Game.fetch_query().where(
            Player.fetch_query().where(
                gameid=Game.id_, username=user.name
            ).exists()
        ).fetchmany()
        ```

        Returns:
            Block[Bool]: The subquery.
        """

        return sql(raw("EXISTS"), wrap(self._get_block()))

    async def fetchmany(
        self, limit: int | None = None
    ) -> LazyList[dict[str, Any], _T]:
        """Execute the query and return a list of models.

        Args:
            limit (int, optional): The maximum number of models to return.
            Defaults to None.

        Raises:
            TypeError: You specified a limit that wasn't an integer (must be a
            python int).

        Returns:
            LazyList[dict, Model]: The list of models matching the query.
        """

        if not (limit is None or isinstance(limit, int)):
            # NOTE: although limit as a string would work, there is a good
            # chance that it's a string because it was user input, meaning
            # that allowing limit to be a string would create an SQL-injection
            # vulnerability.
            raise TypeError("Limit can only be an int.")
        res = await self.con.fetchmany(*self._get_block(limit).render())
        return LazyList(res, _dict_model_converter(self.model))

    async def fetchone(self) -> _T | None:
        """Fetch the first model found.

        Returns:
            Model | None: Returns the model, or None if none were found.
        """

        res = await self.con.fetchrow(*self._get_block().render())
        if res is None:
            return None
        return self.model._from_raw(**res)

    async def count(self) -> int:
        """SELECT COUNT(*) ...

        Returns:
            int: The count.
        """

        return cast(
            int, await self.con.fetchval(*self._get_block(count=True).render())
        )

    async def cursor(self) -> AsyncGenerator[_T, None]:
        """Return an iterator of the resulting models.

        Yields:
            Model: The (next) model from the iterator.
        """

        con = self.con if isinstance(self.con, Connection) else None
        async with self.model.database.cursor(
            *self._get_block().render(), con=con
        ) as cursor:
            async for res in cursor:
                yield self.model._from_raw(**res)

    def _get_block(
        self, limit: int | None = None, count: bool = False
    ) -> Block[Any]:
        if count:
            return select(
                from_=self.model,
                where=self._where_logic(),
                count=True,
                limit=limit,
            )
        return select(
            from_=self.model,
            where=self._where_logic(),
            order_by=self._order_by_logic,
            reverse=self._reverse,
            limit=limit,
        )


class DeleteQueryBuilder(FilterQueryBuilder[_T]):
    """Query builder for deleting models."""

    __slots__: Iterable[str] = ()

    async def execute(self) -> LazyList[dict[str, Any], _T]:
        """Execute the deletion query.

        Returns:
            LazyList[dict, Model]: List of models deleted.
        """

        res = await self.con.fetchmany(*self._get_block().render())
        return LazyList(res, _dict_model_converter(self.model))

    def _get_block(self) -> Block[Any]:
        return delete(
            self.model,
            self._where_logic(),
            list(self.model._all_fields.values()),
        )


class UpdateQueryBuilder(FilterQueryBuilder[_T]):
    """Query builder for updating models."""

    __slots__: Iterable[str] = ("_set_values",)

    def __init__(self, model: Type[_T], con: Connection | None = None) -> None:
        super().__init__(model, con)

        self._set_values: dict[Block[Any], SQL[Any]] = {}

    def set(self, **values: SQL[Any]) -> UpdateQueryBuilder[_T]:
        """Specify changes in the model.

        Example:
        ```
        builder.set(username="New Name")
        ```

        Returns:
            UpdateQueryBuilder: Returns the query builder to allow for
            chaining.
        """

        self._set_values.update({raw(k): v for k, v in values.items()})
        return self

    async def execute(self) -> LazyList[dict[str, Any], _T]:
        """Execute the query.

        Returns:
            LazyList[dict, Model]: List of updated models.
        """

        res = await self.con.fetchmany(*self._get_block().render())
        return LazyList(res, _dict_model_converter(self.model))

    def _get_block(self) -> Block[Any]:
        return update(
            self.model,
            {k: v for k, v in self._set_values.items()},
            where=self._where_logic(),
            return_fields=list(self.model._all_fields.values()),
        )


class InsertQueryBuilder(BaseQueryBuilder[_T]):
    """Query builder for creating a model."""

    __slots__: Iterable[str] = ("_set_values",)

    def __init__(self, model: Type[_T], con: Connection | None = None) -> None:
        super().__init__(model, con)

        self._set_values: dict[Block[Any], SQL[Any]] = {}

    def set(self, **values: SQL[Any]) -> InsertQueryBuilder[_T]:
        """Specify values to be set in the database.

        ```
        await User.insert_query().set(username="Circuit").execute()
        ```

        Returns:
            InsertQueryBuilder: Returns the query builder to allow for
            chaining.
        """

        self._set_values.update({raw(k): v for k, v in values.items()})
        return self

    async def execute(self) -> _T:
        """Execute the query.

        Returns:
            Model: The model that was inserted.
        """

        res = await self.con.fetchrow(*self._get_block().render())
        assert res is not None
        return self.model._from_raw(**res)

    def _get_block(self) -> Block[Any]:
        value_names = list(self._set_values.keys())
        value_values = list(self._set_values.values())

        return insert(
            self.model,
            value_names,
            value_values,
            return_fields=list(self.model._all_fields.values()),
        )



# MIT License
#
# Copyright (c) 2021 TrigonDev
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from typing import Iterable

import asyncpg

from .base_type import SqlType


class Bit(SqlType[asyncpg.BitString]):
    """Fixed length string of 0's and 1's.

    Args:
        length (int, optional): The length of the bitstr. If None,
        postgres uses 1. Defaults to None.

    https://www.postgresql.org/docs/14/datatype-bit.html
    """

    __slots__: Iterable[str] = ("_length",)

    def __init__(self, length: int | None = None) -> None:
        self._length = length
        self._sql = "BIT"
        if length is not None:
            self._sql += f"({length})"

    @property
    def length(self) -> int | None:
        """The length of the bitstr.

        Returns:
            int | None
        """

        return self._length


class VarBit(SqlType[asyncpg.BitString]):
    """Variable-length bitstr with max-length.

    https://www.postgresql.org/docs/14/datatype-bit.html
    """

    def __init__(self, max_length: int | None = None) -> None:
        """Create a VarBit type.

        Args:
            max_length (int, optional): The maximum bitstr length. If None the
            length is unlimited. Defaults to None.
        """

        self._max_length = max_length
        self._sql = "VARBIT"
        if max_length is not None:
            self._sql += f"({max_length})"

    @property
    def max_length(self) -> int | None:
        """The maximum length of the bitstr.

        Returns:
            int | None
        """

        return self._max_length



# MIT License
#
# Copyright (c) 2021 TrigonDev
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Iterable

from apgorm.indexes import IndexType, _IndexType
from apgorm.sql.sql import Block, join, raw, wrap

from .constraint import Constraint

if TYPE_CHECKING:  # pragma: no cover
    from apgorm.field import BaseField
    from apgorm.types.boolean import Bool


class Exclude(Constraint):
    __slots__: Iterable[str] = ("using", "elements", "where")

    def __init__(
        self,
        *elements: tuple[BaseField[Any, Any, Any] | Block[Any] | str, str],
        using: IndexType = IndexType.BTREE,
        where: Block[Bool] | str | None = None,
    ) -> None:
        """Specify an Exclusion constraint for a table.

        Args:
            *elements: key-value pairs of fields and the operator to use on
            that field.

        Kwargs:
            using (IndexType, optional): The index to use. Defaults
            to IndexType.BTREE.
            where (Block[Bool] | str, optional): Specify condition for
            applying this constraint. Defaults to None.
        """

        self.using: _IndexType = using.value
        self.elements = [
            (raw(f) if isinstance(f, str) else f, op) for f, op in elements
        ]
        self.where = raw(where) if isinstance(where, str) else where

    def _creation_sql(self) -> Block[Any]:
        sql = Block[Any](
            raw("CONSTRAINT"),
            raw(self.name),
            raw("EXCLUDE USING"),
            raw(self.using.name),
            join(
                raw(","),
                *(wrap(f, raw("WITH"), raw(op)) for f, op in self.elements),
                wrap=True,
            ),
        )
        if self.where is not None:
            sql += Block(raw("WHERE"), wrap(self.where))

        return sql



# MIT License
#
# Copyright (c) 2021 TrigonDev
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from typing import Any, Coroutine, Iterable

import asyncpg
from asyncpg.cursor import CursorFactory
from asyncpg.transaction import Transaction

from .utils.lazy_list import LazyList


class PoolAcquireContext:
    __slots__: Iterable[str] = ("pac",)

    def __init__(self, pac: asyncpg.pool.PoolAcquireContext) -> None:
        self.pac = pac

    async def __aenter__(self) -> Connection:
        return Connection(await self.pac.__aenter__())

    async def __aexit__(self, *exc: Exception) -> None:
        await self.pac.__aexit__(*exc)


class Pool:
    __slots__: Iterable[str] = ("pool",)

    def __init__(self, pool: asyncpg.Pool) -> None:
        self.pool = pool

    def acquire(self) -> PoolAcquireContext:
        return PoolAcquireContext(self.pool.acquire())

    def close(self) -> Coroutine[Any, Any, None]:
        return self.pool.close()  # type: ignore


class Connection:
    """Wrapper around asyncpg.Connection."""

    __slots__: Iterable[str] = ("con",)

    def __init__(self, con: asyncpg.Connection) -> None:
        self.con = con

    def transaction(self) -> Transaction:
        """Enter a transaction.

        Usage:
        ```
        async with con.transaction():
            await con.execute("DROP TABLE users", [])
            raise Exception("Wait no I don't want to do that!")
        ```

        Alternatively:
        ```
        t = con.transaction()
        await t.start()
        await con.execute("DROP TABLE users", [])
        await t.rollback()

        # or, if you did want to do that, replace t.rollback() with t.commit():
        await t.commit()
        ```

        Returns:
            Transaction: The asyncpg.Transaction object.
        """

        return self.con.transaction()

    async def execute(
        self, query: str, params: list[Any] | None = None
    ) -> None:
        """Execute SQL.

        Consider using Database.execute() unless you want to manage the
        transaction flow.

        Args:
            query (str): The raw SQL.
            params (list[Any], optional): List of parameters. Defaults to None.
        """

        params = params or []
        await self.con.execute(query, *params)

    async def fetchrow(
        self, query: str, params: list[Any] | None = None
    ) -> dict[str, Any] | None:
        """Execute SQL and return a single row, if any.

        Consider using Database.fetchrow() unless you want to manage the
        transaction flow.

        Args:
            query (str): The raw SQL.
            params (list[Any], optional): List of parameters. Defaults to None.

        Returns:
            dict | None: The row or None.
        """

        params = params or []
        res = await self.con.fetchrow(query, *params)
        if res is not None:
            res = dict(res)
        assert res is None or isinstance(res, dict)
        return res

    async def fetchmany(
        self, query: str, params: list[Any] | None = None
    ) -> LazyList[asyncpg.Record, dict[str, Any]]:
        """Execute SQL, returning all found rows.

        See LazyList docs.

        Consider using Database.fetchmany() unless you want to manage the
        transaction flow.

        Args:
            query (str): The raw SQL.
            params (list[Any], optional): List of parameters. Defaults to None.

        Returns:
            LazyList[asyncpg.Record, dict[str, Any]]: [description]
        """

        params = params or []
        return LazyList(await self.con.fetch(query, *params), dict)

    async def fetchval(
        self, query: str, params: list[Any] | None = None
    ) -> Any:
        """Execute SQL, returning the values.

        For example:
        ```
        db.fetchrow("SELECT name FROM users")
        # -> {"name": "Circuit"}

        db.fetchval("SELECT name FROM users")
        # -> "Circuit"
        ```

        Args:
            query (str): [description]
            params (list[Any], optional): [description]. Defaults to None.

        Returns:
            Any: [description]
        """

        return await self.con.fetchval(query, *params)

    def cursor(
        self, query: str, params: list[Any] | None = None
    ) -> CursorFactory:
        """Returns a CursorFactory, but outside a transaction.

        Since cursors must be inside a transaction, you must handle the
        transaction flow like this:

        ```
        async with con.transaction():
            async for row in con.cursor():
                ...
        ```

        Args:
            query (str): The raw SQL
            params (list[Any], optional): List of parameters. Defaults to None.

        Returns:
            CursorFactory: The cursor factory.
        """

        params = params or []
        return self.con.cursor(query, *params)



# MIT License
#
# Copyright (c) 2021 TrigonDev
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
apgorm is a fully type-checked asynchronous ORM wrapped around asyncpg.

apgorm is licensed under the MIT license.

https://github.com/TrigonDev/apgorm
"""

from importlib import metadata

from . import exceptions
from .connection import Connection, Pool, PoolAcquireContext
from .constraints.check import Check
from .constraints.constraint import Constraint
from .constraints.exclude import Exclude
from .constraints.foreign_key import ForeignKey, ForeignKeyAction
from .constraints.primary_key import PrimaryKey
from .constraints.unique import Unique
from .converter import Converter, IntEFConverter
from .database import Database
from .field import BaseField, ConverterField, Field
from .indexes import Index, IndexType
from .manytomany import ManyToMany
from .migrations.describe import (
    Describe,
    DescribeConstraint,
    DescribeField,
    DescribeIndex,
    DescribeTable,
)
from .migrations.migration import Migration
from .model import Model
from .sql.query_builder import (
    BaseQueryBuilder,
    DeleteQueryBuilder,
    FetchQueryBuilder,
    FilterQueryBuilder,
    InsertQueryBuilder,
    UpdateQueryBuilder,
)
from .sql.sql import (
    CASTED,
    SQL,
    Block,
    Comparable,
    Parameter,
    Raw,
    Renderer,
    and_,
    join,
    or_,
    raw,
    sql,
    wrap,
)
from .undefined import UNDEF
from .utils.lazy_list import LazyList

__version__ = metadata.version(__name__)

__all__ = (
    "Converter",
    "Model",
    "Database",
    "ManyToMany",
    "Constraint",
    "Check",
    "ForeignKey",
    "ForeignKeyAction",
    "PrimaryKey",
    "Exclude",
    "Unique",
    "Index",
    "IndexType",
    "Migration",
    "Describe",
    "DescribeConstraint",
    "DescribeTable",
    "DescribeField",
    "DescribeIndex",
    "CASTED",
    "SQL",
    "Block",
    "Comparable",
    "Parameter",
    "Raw",
    "Renderer",
    "BaseQueryBuilder",
    "FetchQueryBuilder",
    "DeleteQueryBuilder",
    "FilterQueryBuilder",
    "InsertQueryBuilder",
    "UpdateQueryBuilder",
    "LazyList",
    "Connection",
    "Pool",
    "Field",
    "BaseField",
    "ConverterField",
    "PoolAcquireContext",
    "IntEFConverter",
    "UNDEF",
    "and_",
    "join",
    "or_",
    "raw",
    "sql",
    "wrap",
    "types",
    "migrations",
    "exceptions",
    "undefined",
)



# MIT License
#
# Copyright (c) 2021 TrigonDev
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from typing import Iterable

import asyncpg

from .base_type import SqlType


class Point(SqlType[asyncpg.Point]):
    """Point (x, y).

    https://www.postgresql.org/docs/14/datatype-geometric.html#id-1.5.7.16.5
    """

    __slots__: Iterable[str] = ()

    _sql = "POINT"


class Line(SqlType[asyncpg.Line]):
    """Infinite line.

    https://www.postgresql.org/docs/14/datatype-geometric.html#DATATYPE-LINE
    """

    __slots__: Iterable[str] = ()

    _sql = "LINE"


class LineSegment(SqlType[asyncpg.LineSegment]):
    """Finite line.

    https://www.postgresql.org/docs/14/datatype-geometric.html#DATATYPE-LSEG
    """

    __slots__: Iterable[str] = ()

    _sql = "LSEG"


class Box(SqlType[asyncpg.Box]):
    """Rectangular box (two points specifying corners).

    https://www.postgresql.org/docs/14/datatype-geometric.html#id-1.5.7.16.8
    """

    __slots__: Iterable[str] = ()

    _sql = "BOX"


class Path(SqlType[asyncpg.Path]):
    """Path (multiple LineSegments). Can be open or closed. A closed path is
    similar to a polygon.

    https://www.postgresql.org/docs/14/datatype-geometric.html#id-1.5.7.16.9
    """

    __slots__: Iterable[str] = ()

    _sql = "PATH"


class Polygon(SqlType[asyncpg.Polygon]):
    """Polygon type (similar to a closed path).

    https://www.postgresql.org/docs/14/datatype-geometric.html#DATATYPE-POLYGON
    """

    __slots__: Iterable[str] = ()

    _sql = "POLYGON"


class Circle(SqlType[asyncpg.Circle]):
    """Circle type (center Point and radius).

    https://www.postgresql.org/docs/14/datatype-geometric.html#DATATYPE-CIRCLE
    """

    __slots__: Iterable[str] = ()

    _sql = "CIRCLE"



# MIT License
#
# Copyright (c) 2021 TrigonDev
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

import ipaddress as ipaddr
from typing import Iterable, Union

from .base_type import SqlType


class CIDR(SqlType[Union[ipaddr.IPv4Network, ipaddr.IPv6Network]]):
    """IPv4 and IPv6 networks. 7 or 19 bytes.

    https://www.postgresql.org/docs/14/datatype-net-types.html#DATATYPE-CIDR
    https://www.postgresql.org/docs/14/datatype-net-types.html#DATATYPE-INET-VS-CIDR
    """

    __slots__: Iterable[str] = ()

    _sql = "CIDR"


class INET(
    SqlType[
        Union[
            ipaddr.IPv6Address,
            ipaddr.IPv6Interface,
            ipaddr.IPv4Address,
            ipaddr.IPv4Interface,
        ]
    ]
):
    """IPv4 and IPv6 hosts and networks. 7 or 19 bytes.

    https://www.postgresql.org/docs/14/datatype-net-types.html#DATATYPE-INET
    https://www.postgresql.org/docs/14/datatype-net-types.html#DATATYPE-INET-VS-CIDR
    """

    __slots__: Iterable[str] = ()

    _sql = "INET"


class MacAddr(SqlType[str]):
    """MAC addresses. 6 bytes.

    https://www.postgresql.org/docs/14/datatype-net-types.html#DATATYPE-MACADDR
    """

    __slots__: Iterable[str] = ()

    _sql = "MACADDR"


class MacAddr8(SqlType[str]):
    """MAC address (EUI-64 format). 8 bytes.

    https://www.postgresql.org/docs/14/datatype-net-types.html#DATATYPE-MACADDR8
    """

    __slots__: Iterable[str] = ()

    _sql = "MACADDR8"
